# FILE: mmarchive.tcl
#
# Accept data from solvers and save to file.
# Accept both vectorfield and datatable formats
# Provide no GUI, so this app survives death of X

# Support libraries
package require Oc
package require Net
if {[Oc_Main HasTk]} {
    wm withdraw .
}

Oc_Main SetAppName mmArchive
Oc_Main SetVersion 1.2.0.0
regexp \\\044Date:(.*)\\\044 {$Date: 2000/04/06 06:34:02 $} _ date
Oc_Main SetDate [string trim $date]
Oc_Main SetAuthor [Oc_Person Lookup dgp]
Oc_CommandLine Parse $argv

##------ Control server ---------#
#
Net_Protocol New protocol(ctrl) -name "OOMMF remotecontrol protocol 0.0"
$protocol(ctrl) AddMessage start HasGui {} {
    Oc_EventHandler New _ all Status \
	    [list $connection Tell Status %message] \
	    -groups [list $connection]
    return [list start [list 0 1]]
}
$protocol(ctrl) AddMessage start Commands {} {
    return [list start [list 0 {}]]
}
$protocol(ctrl) AddMessage start Outputs {} {
    return [list start [list 0 {}]]
}
$protocol(ctrl) AddMessage start Inputs {} {
    return [list start [list 0 {}]]
}
$protocol(ctrl) AddMessage start Events {} {
    return [list start [list 0 {}]]
}
$protocol(ctrl) AddMessage start delete {} {
    return [list start [list 0 [StopServers]]]
}
Net_Server New server(ctrl) -protocol $protocol(ctrl) \
	-alias [Oc_Main GetAppName]
#
##------ Control server ---------#

##------ DataTable server ---------#
#
Net_Protocol New protocol(dt) -name "OOMMF DataTable protocol 0.0"
$protocol(dt) AddMessage start HasGui {} {
    return [list start [list 0 0]]
}
$protocol(dt) AddMessage start DataTable { triples } {
    WriteTriples $connection $triples
    return [list start [list 0 0]]
}
Net_Server New server(dt) -protocol $protocol(dt) \
	-alias [Oc_Main GetAppName]
#
##------ DataTable server ---------#

##------ Vectorfield server ---------#
#
Net_Protocol New protocol(vf) -name "OOMMF vectorField protocol 0.0"
$protocol(vf) AddMessage start datafile { fnlist } {
    set tmpfile [lindex $fnlist 0]
    set permfile [lindex $fnlist 1]
    # Report writing of file
    # Ought to add lots of error checking here...
    Oc_EventHandler Generate all Status -message "Writing $permfile ..."
    file rename -force $tmpfile $permfile
    return [list start [list 0 0]]
}
$protocol(vf) AddMessage start HasGui {} {
    return [list start [list 0 0]]
}
Net_Server New server(vf) -protocol $protocol(vf) \
	-alias [Oc_Main GetAppName]
#
##------ Vectorfield server ---------#

##------ DataTable helper procs ---------#
proc WriteTriples {conn triples} {
    global conn_file_req   ;# Maps connection to requested filename
    global conn_file       ;# Maps connection to actual filename
    global conn_channel    ;# Maps connection to open channel
    global conn_col_list   ;# Maps connection to table column labels
    global conn_head_width ;# Column header widths
    global conn_data_fmt    ;# Format list for data

    set basewidth 15        ;# Default column width
    set basefmt "%- 15.8g"  ;# Base format for table body

    # Create "tripdata" array that maps labels to values
    foreach trip $triples {
	foreach {label {} value} $trip {
	    set tripdata($label) $value
	}
    }

    # Close current file?
    set newfile [lindex [array names tripdata "@Filename:*"] end]
    if {![string match {} $newfile]} {
	set newfile [string trim [string range $newfile 10 end]]
	if {[string match {} $newfile]} {
	    # Explicit close-file request
	    CloseTriplesFile $conn
	    return
	}
	if {[info exists conn_file_req($conn)] && \
		![string match $conn_file_req($conn) $newfile]} {
	    CloseTriplesFile $conn
	}
    }


    # Have we already opened a file for this connection?
    if {[catch {set chan $conn_channel($conn)}]} {
	# New connection!

	# Build column list
	set conn_file($conn) {}
	set conn_file_req($conn) $newfile
	set conn_channel($conn) {}
	set conn_col_list($conn) {}
        set conn_head_width($conn) {}
        set conn_data_fmt($conn) {}
	set unit_list {}
	foreach trip $triples {
	    foreach {label unit {}} $trip {
		if {![string match "@Filename:*" $label]} {
		    lappend conn_col_list($conn) $label  ;# New data column
		    lappend unit_list $unit
		}
	    }
	}

        # Determine column formats
        foreach label $conn_col_list($conn) unit $unit_list {
            set width [string length [list $label]]
            set unit_width [string length [list $unit]]
            if {$width<$unit_width} { set width $unit_width }
            lappend conn_head_width($conn) $width
            set fmt " $basefmt"
            if {$width>$basewidth} {
                # Extra-wide column.  Pad numeric format with
                # spaces on both sides.
                set left [expr {($width-$basewidth)/2}]
                set right [expr {$width-$basewidth-$left}]
                set fmt [format " %*s%s%*s" $left {} $basefmt $right {}]
            }
            lappend conn_data_fmt($conn) $fmt
        }

	# Open output file for append
	set chan {}
	set conn_file($conn) $conn_file_req($conn)
	if {[string match {} $conn_file($conn)] || \
		[catch {open $conn_file($conn) "a+"} chan]} {
	    # Unable to open output file.  Try using temporary filename
	    set conn_file($conn) [Oc_TempName {} "mmArchive" ".odt"]
	    if {[catch {open $conn_file($conn) "a+"} chan]} {
		# Abandon ship!
		error "Unable to open DataTable output file: $chan"
	    }
	}
        Oc_EventHandler Generate all Status -message "Opened $conn_file($conn)"
	set conn_channel($conn) $chan

	# Setup exit handler
        global server
	Oc_EventHandler New _ $conn delete \
		[list CloseTriplesFile $conn] -oneshot 1
	Oc_EventHandler New _ Oc_Main Exit     \
		[list CloseTriplesFile $conn]  \
		-oneshot 1 -groups [list $conn]

	# Write out table header
	if {[tell $chan]<1} {
	    puts $chan "# ODT 1.0" ;# File magic
	} else {
	    puts $chan "#"         ;# Separate blocks with an empty line
	}
	puts $chan "# Table Start"
	puts $chan "# Title: mmArchive Data Table,\
		[clock format [clock seconds]]"
	puts -nonewline $chan "# Columns:"
        # Drop labels and units on left of each column, mutually centered.
	foreach label $conn_col_list($conn) \
                headwidth $conn_head_width($conn) {
            set key [list $label]
            set keylength [string length $key]
            if {$keylength<$headwidth} {
                set left [expr {($headwidth-$keylength)/2}]
                set right [expr {$headwidth-$keylength-$left}]
            } else {
                set left 0
                set right 0
            }
            if {$headwidth<$basewidth} {
                incr right [expr {$basewidth-$headwidth}]
            }
            puts -nonewline $chan \
                    [format " %*s%s%*s" $left {} $key $right {}]
	}
	puts -nonewline $chan "\n# Units:  "
	foreach unit $unit_list \
                headwidth $conn_head_width($conn) {
            set key [list $unit]
            set keylength [string length $key]
            if {$keylength<$headwidth} {
                set left [expr {($headwidth-$keylength)/2}]
                set right [expr {$headwidth-$keylength-$left}]
            } else {
                set left 0
                set right 0
            }
            if {$headwidth<$basewidth} {
                incr right [expr {$basewidth-$headwidth}]
            }
            puts -nonewline $chan \
                    [format " %*s%s%*s" $left {} $key $right {}]
	}
	puts $chan {}  ;# newline
	flush $chan
    }

    # Index into tripdata using the column list.  Note that
    # there is no check for extra (new) data columns; any
    # such will just be ignored.
    puts -nonewline $chan "          "  ;# Indent past "# Columns:" heading
    foreach label $conn_col_list($conn)  \
            width $conn_head_width($conn) \
            fmt $conn_data_fmt($conn) {
	if {[catch {set tripdata($label)} value]} {
	    # Missing value.  This shouldn't happen, but the format could
            # be pre-computed and saved in a list if needed (for speed).
            set left [expr {($width-2)/2}]
            if {$width<$basewidth} { set width $basewidth }
            set right [expr {$width-2-$left}]
	    set value [format " %*s{}%*s" $left {} $right {}]
	} else {
	    set value [format $fmt $value]
	}
	puts -nonewline $chan $value
    }
    puts $chan {}
    flush $chan
}

proc CloseTriplesFile {conn} {
    global conn_file_req conn_file conn_channel conn_col_list
    global conn_head_width conn_data_fmt
    if {![catch {set conn_channel($conn)} chan]} {
	puts $chan "# Table End"
	close $chan
	Oc_EventHandler Generate all Status \
		-message "Closing $conn_file($conn) ..."
    }
    catch {unset conn_file_req($conn)}
    catch {unset conn_file($conn)}
    catch {unset conn_channel($conn)}
    catch {unset conn_col_list($conn)}
    catch {unset conn_head_width($conn)}
    catch {unset conn_data_fmt($conn)}
}

##^^---- DataTable helper procs -------^^#

proc StartServers {} {
    global server
    foreach s [array names server] {
	$server($s) Start 0
    }
}

proc StopServers {} {
    global server
    foreach s [array names server] {
	$server($s) Delete
    }
    # Wait for deregistation
    after 2000 exit
}

Oc_EventHandler Bindtags all all
StartServers
vwait forever
