/* FILE: director.cc                 -*-Mode: c++-*-
 *
 * 3D solver director class.
 *
 */

#undef REPORT_CONSTRUCT_TIMES  // #def this to get timing report on stderr

#include <algorithm>
#include <vector>
#include <string>
#include "oc.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "ext.h"
#include "energy.h"
#include "util.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"

#ifdef REPORT_CONSTRUCT_TIMES
#include "nb.h"
#endif

/* End includes */


int myEval(Tcl_Interp* interp,const string& cmd)
{
  char* buf=new char[cmd.length()+1];
  strcpy(buf,cmd.c_str());
  int errcode=Tcl_Eval(interp,buf);
  delete[] buf;
  return errcode;
}

// The following two routines make sure that all Oxs_Director
// instances are destroyed as part of a Tcl_Exit().
void Oxs_Director::ExitProc(ClientData clientData)
{
    Oxs_Director *d = (Oxs_Director *) clientData;
    delete d;
}

Oxs_Director::Oxs_Director(Tcl_Interp *i) 
  : problem_count(0),problem_id(0),
    interp(i), safe_interp(NULL), driver(NULL),
    total_energy_output(this,"Director","Total energy","J",1,
			&Oxs_Director::UpdateOutputs),
    total_field_output(this,"Director","Total field","A/m",1,
			&Oxs_Director::UpdateOutputs),
    // MS VC++ 6.0 requires fully qualified member names
    simulation_state_reserve_count(0)
{
  Tcl_CreateExitHandler((Tcl_ExitProc *) ExitProc, (ClientData) this);
}

int Oxs_Director::ProbReset()
{ // Sets micromag problem to initial state
  // Run Init() on all Oxs_Ext's
  vector<Oxs_Ext*>::iterator it = ext_obj.begin();
  while(it!=ext_obj.end()) {
    if(!((*it)->Init())) return TCL_ERROR;
    it++;
  }
  return TCL_OK;
}

int Oxs_Director::ProbRun(Tcl_Interp* interp, int argc, char** argv)
{
  string result;
  int errorcode=1;
  if(driver==NULL) {
    result = "Problem not initialized (no driver identified)";
  } else {
    errorcode = driver->Run(result,argc,argv,interp);
  }

    if (errorcode == 0) {
	return myEval(interp, result);
    }
    Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
    return TCL_ERROR;
}

int Oxs_Director::ProbInit(const char* mif,const char* filename)
{ // Import mif must be the name of an existing MIF object
  // in the Tcl interpreter 'interp' passed to the constructor
  // of *this.
  int i,error_code;
  string cmd;

  Release();
  simulation_state_reserve_count=0;  // Safety

  // Create slave interpreter for mif object to use
  cmd = "interp create -safe";
  error_code=myEval(interp,cmd);
  if(error_code!=TCL_OK) return error_code;
  safe_interp_name = string(Tcl_GetStringResult(interp));
  safe_interp = Tcl_GetSlave(interp,Oc_AutoBuf(safe_interp_name.c_str()));
  if(safe_interp_name.size()==0 || safe_interp==NULL) {
    Tcl_AppendResult(interp,
                     "Unable to create safe interpreter",
		     (char *)NULL);
    return TCL_ERROR;
  }

  // Read file into Oxs_Mif object
  cmd = string(mif) + string(" Read ") + safe_interp_name
    + string(" [list ") + string(filename) + string(" ]");
  error_code=myEval(interp,cmd);
  if(error_code!=TCL_OK) {
    return error_code;
  }

  // Get list of spec blocks in mif object
  cmd = string(mif) + string(" GetSpecKeys");
  error_code=myEval(interp,cmd);
  if(error_code!=TCL_OK) return error_code;

  Oxs_SplitList arglist;
  error_code = arglist.Split(Tcl_GetStringResult(interp));
  if(error_code!=TCL_OK) return error_code;

  // Create Oxs_Ext object for each spec block, storing pointer
  // in ext_obj, and mapping on id in ext_map.
  string bad_specs;
#ifdef REPORT_CONSTRUCT_TIMES
  Nb_StopWatch watch;
  Oc_TimeVal cpu,wall;
  fprintf(stderr,
	  "#    TIMING (seconds)               "
	  "                CPU        Elapsed\n");
#endif
  for(i=0;i<arglist.Count();i++) {
    cmd = string(mif) + string(" GetSpecValue [list ")
      + string(arglist[i]) + string(" ]");
    error_code=myEval(interp,cmd);
    if(error_code!=TCL_OK) return error_code;
    string argstr = Tcl_GetStringResult(interp);

    Oxs_Ext* spec_ptr=NULL;
    try {
#ifdef REPORT_CONSTRUCT_TIMES
      watch.Reset();
      watch.Start();
#endif
      spec_ptr=Oxs_Ext::MakeNew(arglist[i],this,
				safe_interp,argstr.c_str());
#ifdef REPORT_CONSTRUCT_TIMES
      watch.Stop();
      watch.GetTimes(cpu,wall);
      fprintf(stderr,"# %40s  %12.3f %12.3f\n",
	      arglist[i],double(cpu),double(wall));
      
#endif
    }
    catch (Oxs_Ext::Error& err) {
      Release();  // Delete any Oxs_Ext objects already constructed
      Tcl_AppendResult(interp,"Oxs_Ext initialization error---",
		       err.msg,(char *)NULL);
      if(bad_specs.length()>0) {
	Tcl_AppendResult(interp,"\nUnrecognized Specify Blocks---",
			 bad_specs.c_str(),(char *)NULL);
      }
      return TCL_ERROR;
    }

    if(spec_ptr!=NULL) {
      string idname=spec_ptr->InstanceName();
      if(ext_map.find(idname)!=ext_map.end()) {
	// Name already in use.  Append to bad spec list and delete.
	delete spec_ptr;
	bad_specs += string("\n ") + string(arglist[i])
	  + string(" (ID name ") + idname + string("already in use.)");
      } else {
	ext_map[idname]=spec_ptr;  // Store mapping
	// Store in ordered list
	ext_obj.push_back(spec_ptr);

	// If this is an Oxs_Energy object, then append it to
	// the energy_obj list as well.
	Oxs_Energy* energy_ptr
	  = dynamic_cast<Oxs_Energy*>(spec_ptr);
	if(energy_ptr!=NULL)
	  energy_obj.push_back(energy_ptr);

	// Set driver data member if this is the first driver object
	if(driver==NULL) {
	  Oxs_Driver* driver_ptr
	    = dynamic_cast<Oxs_Driver*>(spec_ptr);
	  if(driver_ptr!=NULL) driver=driver_ptr;
	}

      }
    } else {
      bad_specs += string("\n  ") + string(arglist[i]);
    }
  }

  // Checks for invalid input
  Tcl_ResetResult(interp);
  if(bad_specs.length()>0 || driver==NULL) {
    Release();  // Delete all Oxs_Ext objects
    if(bad_specs.length()>0) {
      Tcl_AppendResult(interp,"Unrecognized Specify Blocks---",
		       bad_specs.c_str(),"\n",(char *)NULL);
    }
    if(driver==NULL) {
      Tcl_AppendResult(interp,"No driver specified\n",(char *)NULL);
    }
    return TCL_ERROR;
  }

  // Register director outputs
  RegisterOutput(&total_energy_output);
  RegisterOutput(&total_field_output);

  // Run Init() on all Oxs_Ext's
  ProbReset();
  problem_id = ++problem_count;
  return TCL_OK;
}

// The following file-local not_null function is used with find_if
// on output_obj to determine if there are any registered output
// objects.
static bool not_null(const Oxs_Output* obj) { return obj!=NULL; }

void Oxs_Director::Release()
{
  problem_id=0;

  // Release director scratch space
  temp_energy.Release();
  temp_field.Release();

  // Deregister director outputs as necessary
  if(IsRegisteredOutput(&total_energy_output)) {
    DeregisterOutput(&total_energy_output);
  }
  if(IsRegisteredOutput(&total_field_output)) {
    DeregisterOutput(&total_field_output);
  }

  driver=NULL;  // Make sure we don't use this pointer anymore.

  // Release all Oxs_Ext's, from the back forward.  The
  // order is important so that objects at the back of the
  // list can release any dependency locks held on objects
  // at the front of the list.
  vector<Oxs_Ext*>::reverse_iterator ext_it = ext_obj.rbegin();
  while(ext_it!=ext_obj.rend()) {
    delete (*ext_it);
    ++ext_it;
  }
  // Release *all* pointers to these objects.
  ext_obj.clear();
  ext_map.clear();
  energy_obj.clear();

  // Release simulation states.  We don't care about the order.
  vector<Oxs_SimState*>::iterator sim_it = simulation_state.begin();
  while(sim_it!=simulation_state.end()) {
    delete (*sim_it);
    ++sim_it;
  }
  simulation_state.clear();
  simulation_state_reserve_count=0;

  // Release safe slave interpreter
  if(safe_interp!=NULL) {
    string cmd("interp delete ");
    cmd += safe_interp_name;
    myEval(interp,cmd);
    safe_interp=NULL;
    safe_interp_name="";
  }

  // Check that all output objects have deregistered
  vector<Oxs_Output*>::const_iterator output_it
    = find_if(output_obj.begin(),output_obj.end(),not_null);
  if(output_it!=output_obj.end()) {
    OC_THROW(Oxs_BadResourceDealloc,
	     "Not all output objects deregistered upon problem release.");
  }
  output_obj.clear(); // Reset output object registration list

}

Oxs_Director::~Oxs_Director()
{
  Release();
  Tcl_DeleteExitHandler((Tcl_ExitProc *) ExitProc, (ClientData) this);
}

Oxs_Ext*
Oxs_Director::FindExtObject(const string& id) const
{ // Tries to find a match of id against entries in ext_map.
  // If "id" does not include a ":" separator, then it is
  // is compared against the front portion of each key.  If the
  // first character of "id" is ":", then it is compared
  // against the back portion of each key.  Otherwise, "id"
  // must match exactly against the key.  Returns an invalid
  // key if a unique match is not obtained.

  UINT4m match_count=0;
  Oxs_Ext* extobj=NULL;
  map<string,Oxs_Ext*>::const_iterator it;

  enum MATCH_TYPES { FRONT, BACK } match_type;
  if(id[0]==':')                      match_type = BACK;
  else if(id.find(':')==string::npos) match_type = FRONT;
  else {
    // Full exact match required.
    // Note: We use ext_map.find(id) rather than
    // ext_map[id] because the latter will insert a
    // dummy element into the map if 'id' is not in
    // the map.
    it = ext_map.find(id);
    if( it == ext_map.end() )
      return NULL; // No match
    return it->second;
  }

  it = ext_map.begin();
  while(it!=ext_map.end()) {
    const string& test_key = it->first;
    int matched=0;
    switch(match_type) {
      case FRONT: {
	// STL defines substr versions of string.compare(), but
	// they seem to be missing from some implementations.
	string front = test_key.substr(0,id.size());
	if(front.compare(id)==0) {
	  matched=1;
	}
	break;
      }
      case BACK: {
	// STL defines substr versions of string.compare(), but
	// they seem to be missing from some implementations.
	string::size_type index=test_key.find(':');
	if(index!=string::npos) {
	  string back = test_key.substr(index);
	  if(back.compare(id)==0) {
	    matched=1;
	  }
	}
	break;
      }
    }
    if(matched) {
      if(match_count==0) extobj = it->second;
      ++match_count;
    }
    ++it;
  }
  if(match_count==1) return extobj;
  return NULL;
}


string Oxs_Director::ListExtObjects() const
{
  string result;
  vector<Oxs_Ext*>::const_iterator it = ext_obj.begin();
  if(it!=ext_obj.end()) 
    result.append((*it)->InstanceName());
  while((++it)!=ext_obj.end()) {
    result.append("\n");
    result.append((*it)->InstanceName());
  }
  return result;
}

string Oxs_Director::ListEnergyObjects() const
{
  string result;
  vector<Oxs_Energy*>::const_iterator it = energy_obj.begin();
  if(it!=energy_obj.end()) 
    result.append((*it)->InstanceName());
  while((++it)!=energy_obj.end()) {
    result.append("\n");
    result.append((*it)->InstanceName());
  }
  return result;
}

// Routines to register and deregister simulation outputs.
// NOTE: All objects registering output are expected to
// automatically deregister as a consequence of the object
// destructions occuring in Oxs_Director::Release().
void Oxs_Director::RegisterOutput(Oxs_Output* obj)
{
  // Check that object not already registered
  vector<Oxs_Output*>::iterator it
    = find(output_obj.begin(),output_obj.end(),obj);
  if(it != output_obj.end()) {
    // Object already registered
    string msg = string("Output object <")
      + obj->OwnerName() + string(",") 
      + obj->OutputName() + string("> already registered.");
    OC_THROW(Oxs_BadIndex,msg.c_str());
  }

  // Otherwise, append to end of output_obj list
  output_obj.push_back(obj);
}

void Oxs_Director::DeregisterOutput(const Oxs_Output* obj)
{
  vector<Oxs_Output*>::iterator it
    = find(output_obj.begin(),output_obj.end(),obj);
  if(it == output_obj.end()) {
    OC_THROW(Oxs_BadIndex,"Attempted deregistration of "
	     "non-registered output object.");
  }
  *it = NULL;
}

BOOL Oxs_Director::IsRegisteredOutput(const Oxs_Output* obj) const
{
  vector<Oxs_Output*>::const_iterator it
    = find(output_obj.begin(),output_obj.end(),obj);
  if(it == output_obj.end()) return 0;
  return 1;
}

const char*
Oxs_Director::MakeOutputToken(unsigned long index,unsigned long probid) const
{
  static Oc_AutoBuf buf;
  buf.SetLength(4*sizeof(unsigned long)+8);
  sprintf(static_cast<char *>(buf),"%lX %lX",index,probid);
  return static_cast<char *>(buf);
}

Oxs_Output*
Oxs_Director::InterpretOutputToken(const char* token) const
{
  unsigned long index=0,probid=0;
  char* endptr;
  index = strtol(token,&endptr,16);
  if(endptr!=NULL) {
    probid = strtol(endptr+1,&endptr,16);
  }
  if(endptr==NULL) {
    OC_THROW(Oxs_BadParameter,
	     "Invalid token passed to"
	     " Oxs_Director::InterpretOutputToken");
  }
  if(probid!=problem_id) {
    OC_THROW(Oxs_BadParameter,
	     "Problem ID mismatch on token passed to"
	     " Oxs_Director::InterpretOutputToken");
  }
  if(index>=output_obj.size()) {
    OC_THROW(Oxs_BadIndex,
	     "Out-of-range index in token passed to"
	     " Oxs_Director::InterpretOutputToken");
  }
  Oxs_Output* obj = output_obj[index];
  if(obj==NULL) {
    OC_THROW(Oxs_BadParameter,
	     "Token refers to deregistered output object in"
	     " Oxs_Director::InterpretOutputToken");
  }
  return obj;
}

void Oxs_Director::ListOutputObjects(vector<string> &outputs) const
{
  outputs.clear();
  size_t size = output_obj.size();
  for(size_t i=0;i<size;i++) {
    const Oxs_Output* obj = output_obj[i];
    if(obj!=NULL) {
      outputs.push_back(MakeOutputToken(i,problem_id));
    }
  }
}

int Oxs_Director::Output(const char* output_token,
			 Tcl_Interp* interp,
			 int argc,char** argv)
{
  if(driver==NULL) {
    Tcl_AppendResult(interp,"No driver specified\n",(char *)NULL);
    return TCL_ERROR;
  }
  const Oxs_SimState* state=driver->GetCurrentState();
  if(state==NULL) {
    Tcl_AppendResult(interp,"Current state not initialized\n",
		     (char *)NULL);
    return TCL_ERROR;
  }
  Oxs_Output* obj = InterpretOutputToken(output_token);
  return obj->Output(state,interp,argc,argv);
}

int
Oxs_Director::OutputCacheRequestIncrement(const char* output_token,
					  INT4m incr) const
{
  Oxs_Output* obj = InterpretOutputToken(output_token);
  if (obj->CacheRequestIncrement(incr)) return TCL_OK;
  return TCL_ERROR;
}

int Oxs_Director::OutputNames(const char* output_token,
       string& owner_name,string& output_name,
       string& output_type,string& output_units) const
{
  Oxs_Output* obj = InterpretOutputToken(output_token);
  owner_name   = obj->OwnerName();
  output_name  = obj->OutputName();
  output_type  = obj->OutputType();
  output_units = obj->OutputUnits();
  return TCL_OK;
}


// GetTotalEnergyAndmxH: Note that the return is mxH, as opposed to MxH.
// This relieves this routine from needing to know what Ms is, and saves
// an unneeded multiplication (since the evolver is just going to divide
// it back out again to calculate dm/dt (as opposed again to dM/dt)).
// The returned energy array is average energy density for the
// corresponding cell in J/m^3; mxH is in A/m.
void Oxs_Director::GetTotalEnergyAndmxH
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& mxH
 )
{
  // Allocate scratch space as needed
  temp_energy.AdjustSize(state.mesh);
  temp_field.AdjustSize(state.mesh);
  Oxs_MeshValue<ThreeVector>* field_cache = NULL;
  if(total_field_output.GetCacheRequestCount()>0) {
    total_field_output.cache.state_id=0;
    field_cache = &total_field_output.cache.value;
    field_cache->AdjustSize(state.mesh);
  }
  const Oxs_MeshValue<REAL8m>* temp_energy_used=NULL;
  const Oxs_MeshValue<ThreeVector>* temp_field_used=NULL;

  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  /// For convenience

  UINT4m i;
  UINT4m vecsize = state.mesh->Size();
  vector<Oxs_Energy*>::const_iterator it = energy_obj.begin();
  if(it!=energy_obj.end()) {
    (*it)->GetEnergyAndField(state,
			     temp_energy,temp_energy_used,
			     temp_field,temp_field_used);
    if(field_cache!=NULL) {
      *field_cache = *temp_field_used;
    }
    for(i=0;i<vecsize;++i) energy[i] = (*temp_energy_used)[i];
    for(i=0;i<vecsize;++i) {
      mxH[i] = -1 * (*temp_field_used)[i]; // mxH = -Hxm
      mxH[i] ^= spin[i];
    }
    while((++it)!=energy_obj.end()) {
      (*it)->GetEnergyAndField(state,
			       temp_energy,temp_energy_used,
			       temp_field,temp_field_used);
      if(field_cache!=NULL) {
	*field_cache += *temp_field_used;
      }
      for(i=0;i<vecsize;++i) energy[i] += (*temp_energy_used)[i];
      ThreeVector torque;
      for(i=0;i<vecsize;++i) {
	torque = spin[i];
	torque ^= (*temp_field_used)[i];
	mxH[i] += torque;
      }
    }
  } else {
    // No energy terms!
    for(i=0;i<vecsize;++i) {
      energy[i] = 0.0;
      mxH[i].Set(0.,0.,0.);
    }
    if(field_cache!=NULL) {
      *field_cache = ThreeVector(0.,0.,0.);
    }
  }

  if(field_cache!=NULL) {
    total_field_output.cache.state_id=state.Id();
    field_cache=NULL; // Safety
  }

  // Calculate and store total energy sum if output object
  // total_energy_output has cache enabled.
  if (total_energy_output.GetCacheRequestCount()>0) {
    const Oxs_Mesh* mesh = state.mesh;
    REAL8m energy_sum = 0.0;
    for(i=0;i<vecsize;++i) {
      energy_sum += energy[i]*mesh->Volume(i);
    }
    total_energy_output.cache.state_id=state.Id();
    total_energy_output.cache.value=energy_sum;
  }

}

void Oxs_Director::UpdateOutputs(const Oxs_SimState& state)
{
  if(state.Id()==0) { // Safety
    return;
  }

  Oxs_MeshValue<REAL8m> energy(state.mesh);
  Oxs_MeshValue<ThreeVector> mxH(state.mesh);

  GetTotalEnergyAndmxH(state,energy,mxH);
}


// SIMULATION STATE MANAGEMENT: During construction, each Oxs_Ext object
// that wants to create and/or hold simulation states should register
// its intent with the director using ReserveSimulationStateRequest().
// The states are actually constructed only as needed, but the count
// obtained via reserve calls is used as an upper bound to protect
// against memory leaks swallowing up all of system memory.  Objects
// obtain keys (handles) to new states with the GetNewSimulationState()
// call.  The key initially holds a write lock on the object.  The
// caller can change the lock state, make copies, etc., as desired.  The
// GetNewSimulationState() routine scans through the list of states held
// in the simulation_state array, looking for a pre-created state that
// has no locks on it.  If one is found, a key is returned for that
// state.  If all existing states in the simulation_state array are
// locked, then a new state is created (subject to the max state count
// obtained from ReserveSimulationStateRequest() calls).  In all cases
// the returned key holds a write lock on the state.  The states are
// actually deleted only inside the director Release() command, after
// destruction of all Oxs_Ext objects.
//   NOTE: The key passed in to GetNewSimulationState() is allowed to
// hold a lock on an Oxs_SimState, *IF* that state was obtained from a
// previous call to GetNewSimulationState().  On entry,
// GetNewSimulationState() releases any lock currently held by the
// import key, but then immediately tests if the corresponding
// Oxs_SimState is then free.  If so, then GetNewSimulationState()
// resets the key to the same object.  Therefore, if the chance is good
// that the input key holds the sole lock on an object, then the lock
// should *not* be released before calling GetNewSimulationState().

void Oxs_Director::ReserveSimulationStateRequest
(UINT4m reserve_count)
{
  simulation_state_reserve_count+=reserve_count;
}

void Oxs_Director::GetNewSimulationState
(Oxs_Key<Oxs_SimState>& write_key)
{
  // Check first to see if import write_key is currently holding a lock.
  // If so, release it, and immediately check to see if that SimState
  // is available.
  const Oxs_SimState* sptr = write_key.GetPtr();
  if(sptr!=NULL) {
    write_key.Release();
    if( sptr->ReadLockCount()==0 && sptr->WriteLockCount()==0 ) {
      write_key.Set(const_cast<Oxs_SimState*>(sptr));
      write_key.GetWriteReference();
      return;
    }
  }

  // Otherwise, go to the simulation_state list, and try to find
  // an unlocked state, working from the back forward.
  vector<Oxs_SimState*>::reverse_iterator it
    = simulation_state.rbegin();
  while(it!=simulation_state.rend()) {
    if((*it)->ReadLockCount()==0 && (*it)->WriteLockCount()==0) {
      write_key.Set(*it);
      write_key.GetWriteReference();
      return;
    }
    ++it;
  }

  // If we get to here, then all existing states have locks
  // on them.  So create a new state.
  if(simulation_state.size()>=simulation_state_reserve_count) {
    OC_THROW(Oxs_BadResourceAlloc,
	     "Simulation state reserve count exhausted.");
  }
  Oxs_SimState* newstate = new Oxs_SimState;
  simulation_state.push_back(newstate);
  write_key.Set(newstate);
  write_key.GetWriteReference();
}
