/* FILE: filevectorfieldinit.h      -*-Mode: c++-*-
 *
 * File vector field initializer, derived from Oxs_VectorFieldInit class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "filevectorfieldinit.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_FileVectorFieldInit);

/* End includes */


// Constructor
Oxs_FileVectorFieldInit::Oxs_FileVectorFieldInit(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorFieldInit(name,newdtr,safe_interp,argstr),
    file_mesh(NULL),set_norm(0),norm(0.0)
{
  // Process arguments.
  const vector<string>* normlist = FindInitValue("norm");
  if(normlist!=NULL) {
    set_norm = 1;
    CheckInitValueParamCount("norm",1);
    norm = Nb_Atof((*normlist)[0].c_str());
    DeleteInitValue("norm");
  }
  string filename=GetStringInitValue("file");
  VerifyAllInitArgsUsed();

  // Create and load Vf_Mesh object with file data
  Vf_FileInput* vffi=Vf_FileInput::NewReader(filename.c_str());
  if(vffi==NULL) {
    string msg=string("Error detected during construction of ")
      + string(InstanceName()) + string("; Error reading input file ")
      + filename;
    throw Oxs_Ext::Error(msg.c_str());
  }
  file_mesh=vffi->NewMesh();
  delete vffi;
  if(file_mesh==NULL) {
    string msg=string("Error detected during construction of ")
      + string(InstanceName()) + string(" on input file ")
      + filename + string("; Unable to create Vf_Mesh object.");
    throw Oxs_Ext::Error(msg.c_str());
  }
}

Oxs_FileVectorFieldInit::~Oxs_FileVectorFieldInit()
{
  if(file_mesh!=NULL) delete file_mesh;
}


BOOL Oxs_FileVectorFieldInit::Init()
{ return 1; }

void
Oxs_FileVectorFieldInit::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<ThreeVector>& array) const
{
  array.AdjustSize(mesh);

  // Variables for file_mesh (Vf_Mesh*) interface
  Nb_Vec3<REAL8> file_pos;  Nb_LocatedVector<REAL8> file_lv;
  file_lv.value.Set(1.,0.,0.); // Safety

  // Variables for mesh (Oxs_Mesh*) interface
  ThreeVector center,value;

  // Determine scaling
  Oxs_Box bbox;
  mesh->GetBoundingBox(bbox);
  REAL8m xmin = bbox.GetMinX();
  REAL8m ymin = bbox.GetMinY();
  REAL8m zmin = bbox.GetMinZ();
  REAL8m xrange = bbox.GetMaxX() - xmin;
  REAL8m yrange = bbox.GetMaxY() - ymin;
  REAL8m zrange = bbox.GetMaxZ() - zmin;

  Nb_BoundingBox<REAL8> file_box;
  file_mesh->GetPreciseRange(file_box);
  Nb_Vec3<REAL8> file_basept;
  file_box.GetMinPt(file_basept);
  REAL8m file_xmin = file_basept.x;
  REAL8m file_ymin = file_basept.y;
  REAL8m file_zmin = file_basept.z;
  REAL8m file_xrange = file_box.GetWidth();
  REAL8m file_yrange = file_box.GetHeight();
  REAL8m file_zrange = file_box.GetDepth();

  ThreeVector scale,offset;
  if(xrange>0.) {
    scale.x = file_xrange/xrange;
    offset.x = file_xmin - xmin*scale.x;
  } else {
    scale.x = 0.0;
    offset.x = file_xmin/2.0;
  }
  if(yrange>0.) {
    scale.y = file_yrange/yrange;
    offset.y = file_ymin - ymin*scale.y;
  } else {
    scale.y = 0.0;
    offset.y = file_ymin/2.0;
  }
  if(zrange>0.) {
    scale.z = file_zrange/zrange;
    offset.z = file_zmin - zmin*scale.z;
  } else {
    scale.z = 0.0;
    offset.z = file_zmin/2.0;
  }

  // Step through range and sample mesh, fitting
  // with zeroth order fit.
  UINT4m size=mesh->Size();
  for(UINT4m i=0;i<size;i++) {
    mesh->Center(i,center);
    center.x *= scale.x; center.y *= scale.y; center.z *= scale.z;
    center += offset;
    file_pos.x=center.x;
    file_pos.y=center.y;
    file_pos.z=center.z;
    file_mesh->FindPreciseClosest(file_pos,file_lv);

    value.x=file_lv.value.x;
    value.y=file_lv.value.y;
    value.z=file_lv.value.z;
    if(set_norm) value.SetMag(norm);
    array[i]=value;
  }
}


