/* FILE: uzeeman.cc            -*-Mode: c++-*-
 *
 * Uniform Zeeman (applied field) energy, derived from Oxs_Energy class.
 *
 */

#include <stdlib.h>
#include <vector>
#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "uzeeman.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UZeeman);

/* End includes */


// Constructor
Oxs_UZeeman::Oxs_UZeeman(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  // Get scale factor for all Hrange input
  REAL8m hscale=1.0;  // Default value
  const vector<string>* vsptr = FindInitValue("Hscale");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("Hscale",1);
    hscale=Nb_Atof((*vsptr)[0].c_str());
    DeleteInitValue("Hscale");
  }

  // Load Hrange data
  vsptr = FindInitValue("Hrange");
  if(vsptr==NULL || vsptr->size()<1) {
    string msg = "MIF input block for \""
      + string(InstanceName())
      + "\" missing specification for parameter \"Hrange\"";
      throw Oxs_Ext::Error(msg.c_str());
  }


  // Build up Happ vector from Hrange data
  UINT4m step_count;
  ThreeVector h0,h1;
  Oxs_SplitList range_list;
  for(UINT4m i=0;i<vsptr->size();i++) {
    range_list.Split((*vsptr)[i].c_str());
    if( range_list.Count() != 7) {
      char buf[512];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error in MIF input block for \"%s\";"
		  " Hrange[%d] value must be a 7 element list"
		  " (actual element count: %d)",
		  InstanceName(),i,range_list.Count());
      throw Oxs_Ext::Error(buf);
    }
    h0.x = Nb_Atof(range_list[0]);
    h0.y = Nb_Atof(range_list[1]);
    h0.z = Nb_Atof(range_list[2]);
    h1.x = Nb_Atof(range_list[3]);
    h1.y = Nb_Atof(range_list[4]);
    h1.z = Nb_Atof(range_list[5]);
    step_count = atoi(range_list[6]);
    h0*=hscale;
    h1*=hscale;
    if(step_count==0 || Happ.size()==0 || Happ.back()!=h0 ) {
      Happ.push_back(h0);
    }
    for(UINT4m j=1;j<=step_count;j++) {
      REAL8m t = REAL8m(j)/REAL8m(step_count);
      Happ.push_back(((1-t)*h0) + (t*h1));
    }
  }
  DeleteInitValue("Hrange");
  VerifyAllInitArgsUsed();
}

Oxs_UZeeman::~Oxs_UZeeman()
{}

BOOL Oxs_UZeeman::Init()
{ return 1; }

void Oxs_UZeeman::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  UINT4m size = state.mesh->Size();

  UINT4m stage_number = state.stage_number;
  UINT4m field_count = static_cast<UINT4m>(Happ.size());

  ThreeVector H(0,0,0);
  if(field_count>0) {
    if(stage_number>=field_count)  H=Happ[field_count-1];
    else                           H=Happ[stage_number];
  }

  UINT4m i;
  for(i=0;i<size;++i) {
    field[i] = H;
  }

  ThreeVector vtemp(H);
  vtemp *= -MU0; // Constant so that
                /// vtemp * spin * Ms = cell energy density
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  for(i=0;i<size;++i) {
    energy[i] = Ms[i]*(vtemp*spin[i]);
  }

}
